package DataRobot;

import java.io.File;
import java.math.BigDecimal;
import java.nio.file.Files;
import java.util.Base64;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;

import org.apache.http.HttpHost;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.util.EntityUtils;

import com.corticon.services.extensions.ICcServiceCalloutExtension;
import com.corticon.services.dataobject.ICcDataObject;
import com.corticon.services.dataobject.ICcDataObjectManager;

public class Prediction implements ICcServiceCalloutExtension {
	/* Corticonエンティティ名 */
	final static String DATAROBOT = "DataRobot";
	final static String DR_INPUT = "dr_input";
	final static String DR_OUTPUT = "dr_output";
	final static String DR_MAPPING = "dr_mapping";
	
	/* Corticon属性名 */
	final static String cc_item_name = "cc_item_name";
	final static String dr_item_name = "dr_item_name";
	final static String image_flg = "image_flg";
	final static String not_request_flg = "not_request_flg";
	
	final static String cc_label = "label";
	final static String cc_value = "value";
	
	final static String DR_URL = "DR_URL";
	final static String DR_AUTH = "DR_AUTH";
	final static String DR_KEY = "DR_KEY";
	final static String PROXY_IP = "PROXY_IP";
	final static String PROXY_PORT = "PROXY_PORT";
	
	/* サービスコールアウト実行 */
	public static void execute(ICcDataObjectManager dataObjMgr, Properties apropServiceCalloutProperties) {
		try {
			Set<ICcDataObject> drSet = dataObjMgr.getEntitiesByName(DATAROBOT);
			if (drSet == null) {
				return;
			}
			for (ICcDataObject dr : drSet) {
				if (dr == null) {
					continue;
				}
				Set<ICcDataObject> drInputSet = dr.getAssociations(DR_INPUT);
				Map<Integer, ICcDataObject> drInputTree = new TreeMap<>();
				convInputSetTree(dataObjMgr, drInputSet, drInputTree);
				
				Set<ICcDataObject> drMappingSet = dr.getAssociations(DR_MAPPING);
				Map<String, ICcDataObject> drMappingTree =new TreeMap<>();
				convMappingSetTree(dataObjMgr, drMappingSet, drMappingTree);
				
				String requestJson = getRequestJson(dataObjMgr, drInputTree, drMappingTree);
				dataObjMgr.postMessage(ICcDataObjectManager.INFO, "送信JSON：" + requestJson, dr);
				String responseJson = executeRequest(dataObjMgr, requestJson, dr);
				dataObjMgr.postMessage(ICcDataObjectManager.INFO, "受信JSON：" + responseJson, dr);
				
				createResposeEntity(dataObjMgr, drInputTree, responseJson);
			}
		} catch (Exception e) {
			dataObjMgr.postMessage(ICcDataObjectManager.VIOLATION, e.toString(), null);
		}
	}
	
	/* dr_inputエンティティのSetをTreeMapに変換 */
	private static void convInputSetTree(ICcDataObjectManager dataObjMgr, Set<ICcDataObject> drInputSet, Map<Integer, ICcDataObject> drInputTree) {
		try {
			if (drInputSet == null) {
				drInputTree = null;
			}
			int i = 0;
			for (ICcDataObject drInput : drInputSet) {
				if (drInput == null) {
					continue;
				}
				drInputTree.put(i, drInput);
				i++;
			}
		} catch(Exception e) {
			drInputTree = null;
			dataObjMgr.postMessage(ICcDataObjectManager.VIOLATION, e.toString(), null);
		}
	}
	
	/* dr_mappingエンティティのSetをTreeMapに変換 */
	private static void convMappingSetTree(ICcDataObjectManager dataObjMgr, Set<ICcDataObject> drMappingSet, Map<String, ICcDataObject> drMappingTree) {
		try {
			if (drMappingSet == null) {
				drMappingTree = null;
			}
			for (ICcDataObject drMapping : drMappingSet) {
				if (drMapping == null) {
					continue;
				}
				drMappingTree.put((String) drMapping.getAttributeValue(cc_item_name), drMapping);
			}
		} catch(Exception e) {
			drMappingTree = null;
			dataObjMgr.postMessage(ICcDataObjectManager.VIOLATION, e.toString(), null);
		}
	}

	/* CorticonデータオブジェクトからリクエストJSONを作成 */
	private static String getRequestJson(ICcDataObjectManager dataObjMgr, Map<Integer, ICcDataObject> drInputTree, Map<String, ICcDataObject> drMappingTree) {
		try {
			if (drInputTree == null) {
				return "";
			}
			// ひとまずただのオブジェクト文字列連結処理（本来は正しくJSON変換したい）
			String requestJson = "[";
			for (Integer rowId : drInputTree.keySet()) {
				if (requestJson.length() > 1) {
					requestJson += ",";
				}
				requestJson += "{" + getRequestJsonAttribute(dataObjMgr, drInputTree.get(rowId), drMappingTree) + "}";
			}
			requestJson += "]";
			
			return requestJson;
		} catch(Exception e) {
			dataObjMgr.postMessage(ICcDataObjectManager.VIOLATION, e.toString(), null);
		}
		return "";
	}
	
	/* Corticonの属性データをJSON配列に変換 */
	private static String getRequestJsonAttribute(ICcDataObjectManager dataObjMgr, ICcDataObject drInput, Map<String, ICcDataObject> drMappingTree) {
		try {
			if (drInput == null) {
				return "";
			}
			String requestJson = "";
			for (String attributeName : drInput.getAttributeNames()) {
				String featureName = attributeName;
				boolean imageFlg = false;
				boolean notRequestFlg = false;
				
				// 属性名を特徴量名に変換、もしマッピング情報その他がなければ属性名そのまま
				if (drMappingTree != null) {
					ICcDataObject mapping = drMappingTree.get(attributeName);
					if (mapping != null) {
						featureName = (String) mapping.getAttributeValue(dr_item_name);
						imageFlg = (boolean) mapping.getAttributeValue(image_flg);
						notRequestFlg = (boolean) mapping.getAttributeValue(not_request_flg);
					}
				}
				if (notRequestFlg) {
					continue;
				}			
				if (requestJson != "") {
					requestJson += ",";
				}
				requestJson += "\"" + featureName + "\":";
				
				// 特徴量に対する値を設定
				if (imageFlg) {
					// ほんまはもうちょっと真面目にエラー処理とかせなあかん
					String image_path = (String) drInput.getAttributeValue(attributeName);
					if (image_path == "") {
						requestJson += "\"\"";
						continue;
					}
					File file = new File(image_path);
					byte[] data = Files.readAllBytes(file.toPath());
					String base64str = Base64.getEncoder().encodeToString(data);
					requestJson += "\"" + base64str + "\"";
				} else {
					// 本来は日付型とかBoolean型も考慮する必要あり。（今回は未実装）
					switch (drInput.getAttributeDataType(attributeName)) {
					case "Long":
						requestJson += "" + (Long) drInput.getAttributeValue(attributeName);
						break;
					case "BigDecimal":
						requestJson += "" + (BigDecimal) drInput.getAttributeValue(attributeName);
						break;
					case "String":
						requestJson += "\"" + (String) drInput.getAttributeValue(attributeName) + "\"";
						break;
					default:
						requestJson += "\"\"";
						break;
					}
				}
			}
			return requestJson;
		} catch(Exception e) {
			dataObjMgr.postMessage(ICcDataObjectManager.VIOLATION, e.toString(), drInput);
		}
		return "";
	}
	
	/* HTTPリクエストを実行 */
	private static String executeRequest(ICcDataObjectManager dataObjMgr, String requestJson, ICcDataObject dr) {
		try {
			if (requestJson == "") {
				return "";
			}
			// DR接続設定を取得
			String drUrl = (String) dr.getAttributeValue(DR_URL);
			String drAuth = (String) dr.getAttributeValue(DR_AUTH);
			String drKey = (String) dr.getAttributeValue(DR_KEY);
			String proxyIp = (String) dr.getAttributeValue(PROXY_IP);
			int proxyPort = ((Long) dr.getAttributeValue(PROXY_PORT)).intValue();
			
			// POST設定
			HttpPost post = new HttpPost(drUrl);
			post.addHeader("Authorization", drAuth);
			post.addHeader("DataRobot-Key", drKey);
			StringEntity content = new StringEntity(requestJson, "UTF-8");
			content.setContentType("application/json; charset=UTF-8");
			post.setEntity(content);
			
			// HTTPクライアント設定 (PROXY設定が必要な場合)
			HttpClient client = null;
			if (proxyIp == "" || proxyPort == 0) {
				client = HttpClientBuilder.create().build();
			} else {
				client = HttpClientBuilder.create().setProxy(new HttpHost(proxyIp, proxyPort)).build();
			}
			// HTTPリクエスト実行
			HttpResponse response = client.execute(post);
			dataObjMgr.postMessage(ICcDataObjectManager.INFO, "HTTPステータス: " + response.getStatusLine().getStatusCode(), dr);
			if (response.getStatusLine().getStatusCode() == 200) {
				return EntityUtils.toString(response.getEntity());
			}
		} catch (Exception e) {
			dataObjMgr.postMessage(ICcDataObjectManager.VIOLATION, e.toString(), dr);
		}
		return "";
	}
	
	/* レスポンスJSONからCorticonデータオブジェクトを作成 */
	private static void createResposeEntity(ICcDataObjectManager dataObjMgr, Map<Integer, ICcDataObject> drInputTree, String responseJson) {
		try {
			if (responseJson == "" || drInputTree == null) {
				return;
			}
			// ひとまずただのJSON文字列分割処理（本来は正しくオブジェクト変換したい）
			String[] predictArr = responseJson.split("\"rowId\"");
			for (Integer rowId : drInputTree.keySet()) {
				String predict = predictArr[rowId + 1];
				
				//dataObjMgr.postMessage(ICcDataObjectManager.INFO, "predict: " + predict, null);
				
				String label = predict.split("\"label\":\"")[1].split("\",")[0];
				String value = predict.split("\"value\":")[1].split("}")[0];
				
				//dataObjMgr.postMessage(ICcDataObjectManager.INFO, "label: " + label, null);
				//dataObjMgr.postMessage(ICcDataObjectManager.INFO, "value: " + value, null);
				
				ICcDataObject drOutput = dataObjMgr.createEntity(DR_OUTPUT);
				drOutput.setAttributeValue(cc_label, label);
				drOutput.setAttributeValue(cc_value, new BigDecimal(value));
				drInputTree.get(rowId).addAssociation(DR_OUTPUT, drOutput);
			}
		} catch(Exception e) {
			dataObjMgr.postMessage(ICcDataObjectManager.VIOLATION, e.toString(), null);
		}
	}
}
